/*----------------------------------------------------------------------------

  Copyright (c) 2017 rafael grompone von gioi <grompone@gmail.com>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.

  ----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "iio.h"

/*----------------------------------------------------------------------------*/
#ifndef FALSE
#define FALSE 0
#endif /* !FALSE */

#ifndef TRUE
#define TRUE 1
#endif /* !TRUE */

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

#define TEST 0

/*----------------------------------------------------------------------------*/
/*
   Fatal error, print a message to standard-error output and exit.
 */
static void error(char * msg)
{
  fprintf(stderr,"Error: %s\n",msg);
  exit(EXIT_FAILURE);
}

/*----------------------------------------------------------------------------*/
/** Memory allocation, print an error and exit if fail.
 */
static void * xmalloc(size_t size)
{
  void * p;
  if( size == 0 ) error("xmalloc: zero size");
  p = malloc(size);
  if( p == NULL ) error("xmalloc: out of memory");
  return p;
}

static void zeros_int(int *p, int size_p){
  for(int i=0;i<size_p;i++){
      p[i] = 0 ;
  }
}

static void zeros_double(double *p, int size_p){
  for(int i=0;i<size_p;i++){
      p[i] = 0 ;
  }
}


/*----------------------------------------------------------------------------*/
/** Function used for computing the median
 */

void swap(int *p,int *q) {
   int t;
   
   t=*p; 
   *p=*q; 
   *q=t;
}

void sort(int a[],int n) { 
   int i,j,temp;

   for(i=0;i<n-1;i++) {
      for(j=0;j<n-i-1;j++) {
         if(a[j]>a[j+1])
            swap(&a[j],&a[j+1]);
      }
   }
}

void swap_double(double *p,double *q) {
   double t;
   t=*p; 
   *p=*q; 
   *q=t;
}


void sort_double(double a[],int n){ 
   int i,j,temp;
   for(i=0;i<n-1;i++) {
      for(j=0;j<n-i-1;j++) {
         if(a[j]>a[j+1])
            swap_double(&a[j],&a[j+1]);
      }
   }
}

/*----------------------------------------------------------------------------*/
/** Function used for computing the median
 */
double threshold_pixel(double pixel, double alpha, int theta){
    if(pixel > theta)
    pixel = alpha * pixel ;
    else
    pixel = (1 - alpha) * pixel ;
    return pixel ;
}

/*----------------------------------------------------------------------------*/
/*
   Computes the logarithm of NFA to base 10.

   NFA = NT.b(n,k,p)
   the return value is log10(NFA)

   n,k,p - binomial parameters.
   logNT - logarithm of Number of Tests
 */
#define TABSIZE 100000
static double log_nfa(int n, int k, double p, double logNT)
{
  static double inv[TABSIZE];   /* table to keep computed inverse values */
  double tolerance = 0.1;       /* an error of 10% in the result is accepted */
  double log1term,term,bin_term,mult_term,bin_tail,err;
  double p_term = p / (1.0-p);
  int i;

  if( n<0 || k<0 || k>n || p<0.0 || p>1.0 )
    error("wrong n, k or p values in nfa()");

  if( n==0 || k==0 ) return logNT;
  if( n==k ) return logNT + (double)n * log10(p);

  log1term = lgamma((double)n+1.0) - lgamma((double)k+1.0)
           - lgamma((double)(n-k)+1.0)
           + (double)k * log(p) + (double)(n-k) * log(1.0-p);

  term = exp(log1term);
  if( term == 0.0 )                        /* the first term is almost zero */
    {
      if( (double)k > (double)n * p )      /* at begining or end of the tail? */
        return log1term / M_LN10 + logNT;  /* end: use just the first term */
      else
        return logNT;                      /* begin: the tail is roughly 1 */
    }

  bin_tail = term;
  for(i=k+1;i<=n;i++)
    {
      bin_term = (double)(n-i+1) * ( i<TABSIZE ?
                   (inv[i] ? inv[i] : (inv[i]=1.0/(double)i)) : 1.0/(double)i );
      mult_term = bin_term * p_term;
      term *= mult_term;
      bin_tail += term;
      if(bin_term<1.0)
        {
          /* when bin_term<1 then mult_term_j<mult_term_i for j>i.
             then, the error on the binomial tail when truncated at
             the i term can be bounded by a geometric serie of form
             term_i * sum mult_term_i^j.                            */
          err = term * ( ( 1.0 - pow( mult_term, (double)(n-i+1) ) ) /
                         (1.0-mult_term) - 1.0 );

          /* one wants an error at most of tolerance*final_result, or:
             tolerance * abs(-log10(bin_tail)-logNT).
             now, the error that can be accepted on bin_tail is
             given by tolerance*final_result divided by the derivative
             of -log10(x) when x=bin_tail. that is:
             tolerance * abs(-log10(bin_tail)-logNT) / (1/bin_tail)
             finally, we truncate the tail if the error is less than:
             tolerance * abs(-log10(bin_tail)-logNT) * bin_tail        */
          if( err < tolerance * fabs(-log10(bin_tail)-logNT) * bin_tail ) break;
        }
    }
  return log10(bin_tail) + logNT;
}

//static double nfa_dummy(int n, int k, double p, double logNT)








/*----------------------------------------------------------------------------*/
/*                                    Main                                    */
/*----------------------------------------------------------------------------*/
int main(int argc, char ** argv)
{
  double * input;
  double * image;
  double * diff_vert;
  double * diff_vert_tmp ;
  double * diff_horz;
  double * diff_horz_tmp ;
  double * cumulated_vert;
  double * cumulated_horz;
  double * cumulated_vert_med;
  double * cumulated_horz_med;
  double * cumulated_vert_med_centered ;
  double * cumulated_horz_med_centered ;
  int a[3] ; 
  int X,Y,C ;
  int x,y,c,i,j,k,Bx,By ;
  int xmax = -1;
  int ymax = -1;
  double max,min,lnfa, quantile ;
  double alpha ;
  int theta ;
  int N, window_sz ; // JPEG supposed block size and window size 

  
  FILE *output_twocol_horz ;
  output_twocol_horz = fopen("./output_twocol_horz.txt","w");
  FILE *output_twocol_vert ;
  output_twocol_vert = fopen("./output_twocol_vert.txt","w");
  
  FILE *output_twocol_horz_med ;
  output_twocol_horz_med = fopen("./output_twocol_horz_med.txt","w");
  FILE *output_twocol_vert_med ;
  output_twocol_vert_med = fopen("./output_twocol_vert_med.txt","w");
  
  FILE *output_twocol_horz_med_centered ;
  output_twocol_horz_med_centered = fopen("./output_twocol_horz_med_centered.txt","w");
  FILE *output_twocol_vert_med_centered ;
  output_twocol_vert_med_centered = fopen("./output_twocol_vert_med_centered.txt","w");
  
  /* read input */
  if( argc != 7 )
    error("use: jpegblocks <block_size> <image> <window_sz> <alpha> <theta> <quantile> ");
  N = atoi(argv[1]);
  window_sz = atoi(argv[3]);
  alpha = atof(argv[4]);
  theta = atoi(argv[5]);
  quantile = atof(argv[6]) ; 
  input = iio_read_image_double_split(argv[2], &X, &Y, &C);
  if(window_sz > MIN(X,Y)) window_sz = MIN(X,Y) ; 

  /* get memory */
  image = (double *) xmalloc( X * Y * sizeof(double) );
  diff_vert = (double *) xmalloc( X * Y * sizeof(double) );
  diff_vert_tmp = (double *) xmalloc( X * Y *sizeof(double) ) ;
  diff_horz = (double *) xmalloc( X * Y * sizeof(double) );
  diff_horz_tmp = (double *) xmalloc( X * Y * sizeof(double) );
  cumulated_vert = (double *) xmalloc( Y * sizeof(double) );
  cumulated_horz = (double *) xmalloc( X * sizeof(double) );
  cumulated_vert_med = (double *) xmalloc( (Y-2) * sizeof(double) );
  cumulated_horz_med = (double *) xmalloc( (X-2) * sizeof(double) );
  cumulated_vert_med_centered = (double *) xmalloc( (Y-1) * sizeof(double) );
  cumulated_horz_med_centered = (double *) xmalloc( (X-1) * sizeof(double) );
  
  zeros_double(cumulated_vert,Y) ;
  zeros_double(cumulated_horz, X) ; 
  zeros_double(cumulated_vert_med, (Y-2)) ;
  zeros_double(cumulated_horz_med, (X-2)) ;
  zeros_double(cumulated_vert_med_centered, (Y-1)) ;
  zeros_double(cumulated_horz_med_centered, (X-1)) ;
  
  
  /* generate a gray level image from the input
     (this is useful when the input has many channels) */
  for(x=0; x<X; x++)
    for(y=0; y<Y; y++)
      {
        image[x+y*X] = 0.0;
        for(c=0; c<C; c++)
          if(c==0)
          image[x+y*X] += 0.229*input[x+y*X+c*X*Y];
          else if(c==1)
          image[x+y*X] += 0.587*input[x+y*X+c*X*Y];
          else if(c==2) 
           image[x+y*X] += 0.114*input[x+y*X+c*X*Y];
      }

  /* compute absolute value of cross-difference */
  for(x=0; x<X-1; x++)
    for(y=0; y<Y-1; y++){
     // diff_vert[x+y*X] = fabs( image[x+(y+1)*X] - image[x+y*X] ); 
      diff_vert[x+y*X] = threshold_pixel( fabs( image[x+(y+1)*X] - image[x+y*X] ) , alpha, theta) ;   
      diff_vert_tmp[x+y*X] = 10*diff_vert[x+y*X] ;
     // diff_horz[x+y*X] = fabs( image[x+1 +y*X] - image[x+y*X] ) ;
      diff_horz[x+y*X] = threshold_pixel( fabs( image[x+1 +y*X] - image[x+y*X] ) , alpha, theta) ; 
      diff_horz_tmp[x+y*X] = 10*diff_horz[x+y*X] ;                      
	}
  iio_write_image_double_split("horz.png",diff_horz_tmp,X,Y,1);
  iio_write_image_double_split("vert.png",diff_vert_tmp,X,Y,1);
  free( (void *) diff_vert_tmp) ;
  free( (void *) diff_horz_tmp) ;
  
  /* compute the sum */
  for(x=0; x<X; x++){
     cumulated_horz[x] = 0 ;
     for(y=0; y<Y; y++)
  	    cumulated_horz[x] += diff_vert[x+y*X] ; 
  }
  
  for(y=0; y<Y; y++){
     cumulated_vert[y] = 0 ;
     for(x=0; x<X; x++)
  	    cumulated_vert[y] += diff_horz[x+y*X] ;
   }

 /* Write the results in a text file */
 for(x=0;x<window_sz;x++){
 if(!TEST)  fprintf(output_twocol_horz,"%d %f\n",x, cumulated_horz[x]) ;
 else     fprintf(output_twocol_horz,"%f ", cumulated_horz[x]) ;
 }
 for(y=0;y<window_sz;y++){
 if(!TEST)   fprintf(output_twocol_vert,"%d %f\n",y, cumulated_vert[y]) ;
 else     fprintf(output_twocol_vert,"%f ", cumulated_vert[y]) ;
 }
 

 /* Median filtering the result and subtracting it from the initial
 sum performed on each direction */
 
   for(y=1; y<Y-1; y++){
   	  for(i=0;i<3;i++) a[i] = cumulated_vert[y+i-1] ; 
      sort(a,3) ;
      cumulated_vert_med[y] = a[1] ;
      cumulated_vert_med_centered[y] = cumulated_vert[y+1] - cumulated_vert_med[y] ; 
      
  }
  for(x=1; x<X-1; x++){
      for(i=0;i<3;i++) a[i] = cumulated_horz[x+i-1] ;       
      sort(a,3) ;
      cumulated_horz_med[x] = a[1] ;
      cumulated_horz_med_centered[x] = cumulated_horz[x+1] - cumulated_horz_med[x] ; 
   }
  
 /* Write the results in a text file */
 for(x=1;x<window_sz;x++){
 if(!TEST){    
       fprintf(output_twocol_horz_med,"%d %f\n",x, cumulated_horz_med[x]) ;
       fprintf(output_twocol_horz_med_centered,"%d %f\n",x, cumulated_horz_med_centered[x]) ;
  }else{
       fprintf(output_twocol_horz_med,"%f ", cumulated_horz_med[x]) ;
       fprintf(output_twocol_horz_med_centered,"%f ", cumulated_horz_med_centered[x]) ;
    }  
 }
 for(y=1;y<window_sz;y++){
 if(!TEST){   
      fprintf(output_twocol_vert_med,"%d %f\n",y, cumulated_vert_med[y]) ; 
      fprintf(output_twocol_vert_med_centered,"%d %f\n",y, cumulated_vert_med_centered[y]) ;
 }else{
      fprintf(output_twocol_vert_med,"%f ", cumulated_vert_med[y]) ; 
      fprintf(output_twocol_vert_med_centered,"%f ", cumulated_vert_med_centered[y]) ;
      }
 }
  
 /* compute the local maxima modulo 8 */
 
 int ncol = (int) floor(((int)X/(int)N)) ;
 if(TEST){
 printf("ncol = %d\n", ncol) ; //horz
 printf("X = %d\n", X) ;
 }
 int nrow = (int) floor(((int)Y/(int)N)) ; 
 if(TEST){
 printf("nrow = %d\n", nrow) ; // vert
 printf("Y = %d\n", Y) ;
 }
 int quantile_horz, quantile_vert ;
 int * maxima_horz ;
 int * maxima_vert ;
 int *hist_horz ; 
 int *hist_vert ; 
 int ind_quant_horz = (int) floor(X*quantile) ;
 if(TEST) printf("ind_quant_horz = %d\n", ind_quant_horz) ;
 int ind_quant_vert = (int) floor(Y*quantile) ;
 if(TEST) printf("ind_quant_vert = %d\n", ind_quant_vert) ;
 int card_V = 0 ; // cardinal of the V set in the report
 int card_H = 0 ; // cardinal of the H set in the report
 double logNT = log(2*N) ;
 
 maxima_horz = (int *) xmalloc( X * sizeof(int) ); // contains 0 or 1 if maximum or not   
 maxima_vert = (int *) xmalloc( Y * sizeof(int) );
 hist_horz = (int *) xmalloc( N * sizeof(int)) ; // containing the number of peaks for each value
 hist_vert = (int *) xmalloc( N * sizeof(int)) ; // modulo 8
 
 /*=================================================================================
 % First Method : we keep only the quantile% (e.g. 0.9) the highest peaks in amplitude. If a point     
 % of the centered difference is a peak, the it is set to one, otherwise it is set
 % to zero. Then we compute the sum modulo 8 (each 1 being considered as vote), and we 
 % compute the NFA for each of the possible positions. 
 ================================================================================= */
  
  /* Creating temporary pointers just for sorting */
  
double * cumulated_horz_med_centered_tmp ; 
double * cumulated_vert_med_centered_tmp ;
cumulated_horz_med_centered_tmp = (double *) xmalloc((X-2)*sizeof(double)) ; 
cumulated_vert_med_centered_tmp = (double *) xmalloc((Y-2)*sizeof(double)) ; 

for(x=1;x<(X-1);x++) cumulated_horz_med_centered_tmp[x] = cumulated_horz_med_centered[x] ;
for(y=1;y<(Y-1);y++) cumulated_vert_med_centered_tmp[y] = cumulated_vert_med_centered[y] ;
 
 
 /* firt compute the quantiles values*/
 sort_double(cumulated_horz_med_centered_tmp,X-1) ; // sort the value to increasing order
 sort_double(cumulated_vert_med_centered_tmp,Y-1) ; // sort the value to increasing order
 quantile_horz = cumulated_horz_med_centered_tmp[ind_quant_horz] ; 
 if(TEST) printf("quantile_horz = %d\n", quantile_horz) ; 
 quantile_vert = cumulated_vert_med_centered_tmp[ind_quant_vert] ;
 if(TEST) printf("quantile_vert = %d\n", quantile_vert) ; 
 
 /* write the result in text files for test purposes */
 
 if(TEST){
 
 FILE *test_horz_order;
 test_horz_order = fopen("./test_horz_order.txt", "w") ;
 for(i=1;i<(X-1);i++) fprintf(test_horz_order, "%f ", cumulated_horz_med_centered_tmp[i]) ; 
 FILE *test_vert_order;
 test_vert_order = fopen("./test_vert_order.txt", "w") ;
 for(i=1;i<(Y-1);i++) fprintf(test_vert_order, "%f ", cumulated_vert_med_centered_tmp[i]) ;
 
 /* free memory */
 free((void *) cumulated_horz_med_centered_tmp) ;
 free((void *) cumulated_vert_med_centered_tmp) ;
 }
 
 /* Say which points are local maxima  */
 
 for(x=0;x<(X-2);x++){
    if(cumulated_horz_med_centered[x+1] > quantile_horz) {
    maxima_horz[x] = 1 ; 
    card_H++ ;
    }
    else maxima_horz[x] = 0 ;
 }

 for(y=0;y<(Y-2);y++){
    if(cumulated_vert_med_centered[y+1] > quantile_vert) {
    maxima_vert[y] = 1 ;
    card_V++ ;
    }
    else maxima_vert[y] = 0 ;
 }
 
 /* write the result in text files for test purposes */
 
 if(TEST){
  
 FILE *test_maxima_horz;
 test_maxima_horz = fopen("./test_maxima_horz.txt", "w") ;
 for(i=1;i<(X-2);i++) fprintf(test_maxima_horz, "%d ", maxima_horz[i]) ; 
 FILE *test_maxima_vert;
 test_maxima_vert = fopen("./test_maxima_vert.txt", "w") ;
 for(i=1;i<(Y-2);i++) fprintf(test_maxima_vert, "%d ", maxima_vert[i]) ;

 }

zeros_int(hist_horz, 8) ; 
zeros_int(hist_vert, 8) ; 

for(x=0;x<(X-2);x++){
    if(maxima_horz[x] == 1 ) (hist_horz[(x+3)%8])++ ;
}

for(y=0;y<(Y-2);y++){
    if(maxima_vert[y] == 1 ) (hist_vert[(y+3)%8])++ ;
}

if(TEST){
printf("card_H = %d\n", card_H) ;
printf("card_V = %d\n", card_V) ;
}

int sum_hist_horz = 0 ;
int sum_hist_vert = 0 ;

for(i=0;i<8;i++){
printf("hist_horz[%d] = %d  hist_vert[%d] = %d\n", i, hist_horz[i], i, hist_vert[i]) ;
sum_hist_horz += hist_horz[i] ;
sum_hist_vert += hist_vert[i] ;
}

printf("sum_hist_horz = %d  sum_hist_vert = %d\n",sum_hist_horz, sum_hist_vert) ;

/* write the histograms into a text file for verification purpose */

if(TEST){

FILE *histogram_horz ;
histogram_horz = fopen("./histogram_horz.txt","w") ;

FILE *histogram_vert ;
histogram_vert = fopen("./histogram_vert.txt", "w") ;

for(i=0;i<8;i++){
  fprintf(histogram_horz,"%d ", hist_horz[i]) ;
  fprintf(histogram_vert,"%d ", hist_vert[i]) ;
//  fprintf(histogram_horz,"%d %d ",i, hist_horz[i]) ;
//  fprintf(histogram_vert,"%d %d ",i, hist_vert[i]) ;
}

  fclose(histogram_horz) ;
  fclose(histogram_vert) ; 
}
/*=================================================================================
% Second Method : we keep only the points that deviates significantly from the mean
% and then compute the NFA for each of the possible positions. 
================================================================================= */
 
 /* Compute the mean and standard deviation */
 
 /*
 double mean_horz, mean_vert, std_horz, std_vert ;
 
 mean_horz = 0 ;
 std_horz = 0 ;
 for(x=0;x<(X-2);x++){
    mean_horz += cumulated_horz_med_centered[x] ;
 }
 mean_horz /= (X-1) ;
 for(x=0;x<(X-2);x++){
 std_horz += pow((cumulated_horz_med_centered[x] - mean_horz),2) ;
 }
 std_horz /= (X-2) ;
 std_horz = sqrt(std_horz) ; 
 
 mean_vert = 0 ;
 std_vert = 0 ;
 for(y=0;y<(Y-2);y++){
    mean_vert += cumulated_vert_med_centered[y] ;
 }
 mean_vert /= (Y-1) ;
 for(y=0;y<(Y-2);y++){
 std_vert += pow((cumulated_vert_med_centered[y] - mean_vert),2) ;
 }
 std_vert /= (Y-2) ;
 std_vert = sqrt(std_vert) ; 
 
 printf("mean_horz = %f and std_horz = %f\n", mean_horz, std_horz) ;
 printf("mean_vert = %f and std_vert = %f\n", mean_vert, std_vert) ;
 
 */
 
/*=================================================================================
% Third Method : consider as peak the maximum over all successive 8 points sets
% and then compute the NFA for each of the possible positions.  
================================================================================= */

/* now compute the NFA */

double * NFA_horz  ;
double * NFA_vert ;
NFA_horz = (double *) xmalloc(N * sizeof(double)) ; 
NFA_vert = (double *) xmalloc(N * sizeof(double)) ; 
xmax = 0 ;
ymax = 0 ;

for(i=0;i<8;i++){
  //    printf("hist_horz[%d] = %d\n", i, hist_horz[i]) ;
    NFA_horz[i] = log_nfa(card_H, hist_horz[i], 0.125, logNT) ;
    if(NFA_horz[i] < NFA_horz[xmax]) xmax = i ; 
    NFA_vert[i] = log_nfa(card_V, hist_vert[i], 0.125, logNT) ;
    if(NFA_vert[i] < NFA_vert[ymax]) ymax = i ; 
    printf("NFA_horz[%d] = %f and NFA_vert[%d] = %f\n", i, NFA_horz[i],i, NFA_vert[i]) ; 
}

if(TEST){
FILE *output_NFA_horz ;
output_NFA_horz = fopen("./output_NFA_horz.txt", "w") ;
FILE *output_NFA_vert ;
output_NFA_vert = fopen("./output_NFA_vert.txt", "w") ;
for(i=0;i<8;i++){
  fprintf(output_NFA_horz,"%f ", NFA_horz[i]) ;
  fprintf(output_NFA_vert,"%f ", NFA_vert[i]) ;
  }
fclose(output_NFA_horz) ; 
fclose(output_NFA_vert) ;
}

if(NFA_horz[xmax] < 0 && NFA_vert[ymax] < 0){
   printf("JPEG grid detected with probability p = %f\n", (1. - pow(10, MIN(NFA_horz[xmax], NFA_vert[ymax]))));
   printf("Starting at point (%d,%d)\n", xmax, ymax); 
} else {
   printf("JPEG compression unlikely, p = %f\n", (1. - pow(10, MIN(NFA_horz[xmax], NFA_vert[ymax]) - logNT)));
}


/* Find the minimum */


 /* free memory */
  free( (void *) input ) ;
  free( (void *) image ) ;
  free( (void *) diff_horz ) ;
  free( (void *) diff_vert ) ;
  free( (void *) cumulated_horz ) ;
  free( (void *) cumulated_vert ) ;
  free( (void *) cumulated_vert_med ) ; 
  free( (void *) cumulated_horz_med ) ; 
  free( (void *) cumulated_vert_med_centered ) ; 
  free( (void *) cumulated_horz_med_centered ) ; 
  free( (void *) maxima_horz ) ; 
  free( (void *) maxima_vert ) ; 
  free( (void *) hist_horz) ;
  free( (void *) hist_vert) ; 
 // free( (void *) NFA_horz) ;
 // free( (void *) NFA_vert) ;
  fclose(output_twocol_horz) ;
  fclose(output_twocol_vert) ;
  fclose(output_twocol_horz_med) ;
  fclose(output_twocol_vert_med) ;
  fclose(output_twocol_horz_med_centered) ;
  fclose(output_twocol_vert_med_centered) ;
//  free( (void *) a ) ; 
  return EXIT_SUCCESS;
}
/*----------------------------------------------------------------------------*/
